

from .types.role import Role as RolePayload
from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from .state import ConnectionState
    from .team import Team

__all__ = ['Role']

class Role:

    __slots__ = (
        '_state',
        'team',
        'id',
        'name',
        '_icon_url',
        '_color',
        '_color2',
        'permissions',
        'priority',
        '_created_at',
        '_updated_at',
        '_is_displayed_separately',
        '_is_self_assignable',
        '_icon_emoji_id',
        'mentionable',
        '_bot_scope'
    )

    def __init__(
        self,
        state: ConnectionState,
        *,
        team: Team,
        data: RolePayload
    ) -> None:
        self._state: ConnectionState = state
        self.team: Team = team
        self._update(data)

    def _update(self, data: RolePayload):
        self.id: str = data['id']
        #self.team
        self.name: str = data['name']

        self._icon_url: Optional[str] = data.get('iconUrl')
        self._color: str = data['color']
        self._color2: Optional[str] = data.get('color2')
        self.permissions: int = data.get('permissions', 0)
        self.priority: int = data.get('priority', 0)
        self._created_at: str = data['createdAt']
        self._updated_at: Optional[str] = data.get('updatedAt')
        self._is_displayed_separately: bool = data.get('isDisplayedSeparately', True)
        self._is_self_assignable: bool = data.get('isSelfAssignable', False)
        self._icon_emoji_id: Optional[str] = data.get('iconEmojiId')
        self.mentionable: bool = data.get('mentionable', True)
        self._bot_scope: Dict[str,str] = data.get('botScope',{})

    def to_dict(self):
        return {
            "id": self.id,
            "teamId": self.team.id,
            "name": self.name,
            "iconUrl": self._icon_url,
            "color": self._color,
            "color2": self._color2,
            "permissions": self.permissions,
            "priority": self.priority,
            "createdAt": self._created_at,
            "updatedAt": self._updated_at,
            "isDisplayedSeparately": self._is_displayed_separately,
            "isSelfAssignable": self._is_self_assignable,
            "iconEmojiId": self._icon_emoji_id,
            "mentionable": self.mentionable,
            "botScope": self._bot_scope
        }

    def __repr__(self) -> str:
        return f"<Role id={self.id} name={self.name} permissions={self.permissions} teamId={self.team.id}>"



"""
Teamly API Wrapper
~~~~~~~~~~~~~~~~~~~

A basic wrapper for the Teamly API.

:copyright: (c) 2025 Fatih Kuloglu
:license: MIT, see LICENSE for more details.

"""

__title__ = 'teamly'
__author__ = 'MrFatihLD'
__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2025 Fatih Kuloglu'
__version__ = '0.1.0a'

from . import (
    abc as abc,
    utils as utils
)
from .announcement import *
from .blog import *
from .cache import *
from .channel import *
from .client import *
from .color import *
from .embed import *
from .enums import *
from .gateway import *
from .http import *
from .member import *
from .message import *
from .reaction import *
from .state import *
from .team import *
from .todo import *
from .user import *
from .logging import setup_logging, enable_debug #noqa: F401 <- ignoring unused import warning

#hide debug logs by default
setup_logging(False)



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import asyncio

from teamly.logging import enable_debug
from teamly.user import ClientUser

from .http import HTTPClient
from .gateway import TeamlyWebSocket
from .state import ConnectionState

from loguru import logger
from typing import (
    Dict,
    List,
    Optional,
    TypeVar,
    Callable,
    Coroutine,
    Any
)

__all__ = (
    'Client',
)

T = TypeVar('T')
Coro = Coroutine[Any,Any, T]
CoroT = TypeVar('CoroT', bound=Callable[..., Coro[Any]])

class _LoopSentinel:
    __slots__ = ()

    def __getattr__(self, attr: str) -> None:
        msg = (
            'loop attribute cannot be accessed in non-async contexts. '
            'Consider using either an asynchronous main function and passing it to asyncio.run or '
            'using asynchronous initialisation hooks such as Client.setup_hook'
        )
        raise AttributeError(msg)

_loop: Any = _LoopSentinel()


class Client:

    def __init__(
        self,
        *,
        enable_debug: bool = False
    ) -> None:
        self.loop: asyncio.AbstractEventLoop = _loop
        self.http: HTTPClient = HTTPClient(self.loop)
        self.ws: TeamlyWebSocket = None #type: ignore
        self._listener: Dict[str,List[str]] = {}

        self.enable_debug: bool = enable_debug

        self._connection: ConnectionState = self._get_state()

    def run(self, token: str) -> None:
        """
        Starts the client using the provided token.

        This method sets up the asyncio event loop, initializes internal components,
        and begins the WebSocket connection. It blocks the main thread until the
        client is stopped or interrupted.

        Parameters:
            token (str): Your Teamly bot token used for authentication.
        """

        logger.debug('Your token is "{}"', token)

        async def runner():
            logger.debug("starting coroutine")

            if self.enable_debug:
                enable_debug()

            async with self:
                await self.start(token)

        try:
            asyncio.run(runner())
        except KeyboardInterrupt:
            pass

    async def start(self, token: str) -> None:
        await self.http.static_login(token)
        await self.connect()

    async def connect(self) -> None:
        try:
            coro = TeamlyWebSocket.from_client(self)
            self.ws = await asyncio.wait_for(coro,timeout=30)
            while True: #temporaly
                await self.ws.poll_event()
                await asyncio.sleep(1)
        except Exception as e:
            logger.error("Exception error: {}",e)
            await self.close()

    def event(self, coro: CoroT, /) -> CoroT:
        """
        Registers a coroutine function as an event handler.

        Use this as a decorator to bind custom event listeners like `on_ready`,
        `on_message`, etc. The method name must start with `on_` to be recognized.

        Example:
            @client.event
            async def on_ready():
                print("Bot is ready!")

        Parameters:
            coro (Callable): The coroutine function to register.

        Returns:
            Callable: The same coroutine function.

        Raises:
            TypeError: If the function is not a coroutine.
        """

        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("event registered must be a coroutine function")

        setattr(self, coro.__name__, coro)
        logger.debug("\"{}\" has successfully been registered as an event",coro.__name__)
        return coro

    async def _run_event(
        self,
        coro: Callable[...,Coroutine[Any,Any,Any]],
        event_name: str,
        *arg: Any,
        **kwargs: Any
    ) -> None:
        try:
            await coro(*arg, **kwargs)
        except Exception as e:
            logger.error("The event could not run {}",e)

    def _schedul_event(
        self,
        coro: Callable[..., Coroutine[Any,Any,Any]],
        event_name: str,
        *arg,
        **kwargs
    ) -> asyncio.Task:
        wrapper = self._run_event(coro, event_name, *arg, **kwargs)
        #scheduls event
        return self.loop.create_task(wrapper, name=f"Teamly.py: {event_name}")

    def dispatch(self, event: str, /, *arg: Any, **kwargs: Any):
        method = 'on_' + event

        try:
            coro = getattr(self,method)
        except AttributeError:
            pass
        else:
            self._schedul_event(coro, method, *arg, **kwargs)


    async def _async_setup_hook(self):
        logger.debug("setup event loop...")
        loop = asyncio.get_running_loop()
        self.loop = loop
        self.http.loop = loop

    async def close(self):
        """
        Closes the WebSocket and HTTP connections.

        This method should be used when you want to shut down the client cleanly,
        such as during application shutdown or error handling.
        """
        if self.ws is not None:
            await self.ws.close()
        await self.http.close()

    def _get_state(self):
        return ConnectionState(dispatch=self.dispatch,http=self.http)

    @property
    def user(self) -> Optional[ClientUser]:
        """
        Returns the currently authenticated bot user, if available.

        This property becomes available after a successful connection
        and login with a valid token.

        Returns:
            Optional[ClientUser]: The bot's user object.
        """
        return self._connection._user

    async def __aenter__(self):
        """
        Enters the asynchronous context manager.

        This prepares the internal event loop and HTTP client for usage.
        Called automatically when using `async with`.
        """

        logger.debug("Entering context...")
        await self._async_setup_hook()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """
          Exits the asynchronous context manager.

          This ensures the client is closed gracefully when used in an `async with` block.
        """

        logger.debug("Exiting context...")
        await self.close()



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from typing import Optional, Dict

from .color import Color


class Embed:

    __slots__ = (
        "title",
        "description",
        "url",
        "color",
        "_author",
        "_thumbnail",
        "_image",
        "_footer",
    )

    def __init__(
        self,
        *,
        title: Optional[str] = None,
        description: Optional[str] = None,
        url: Optional[str] = None,
        color: Optional[Color] = None
    ):
        if title and len(title) > 16:
            raise ValueError("title must be 16 characters or fewer")
        if description and len(description) > 1024:
            raise ValueError("description must be 1024 characters or fewer")

        self.title = title
        self.description = description
        self.url = url
        self.color = int(color) if color else None

        self._author: Optional[Dict[str, str]] = None
        self._thumbnail: Optional[Dict[str, str]] = None
        self._image: Optional[Dict[str, str]] = None
        self._footer: Optional[Dict[str, str]] = None

    @property
    def author(self) -> Dict[str,str] | None:
        return self._author if self._author else None

    @property
    def thumbnail(self) -> str | None:
        return self._thumbnail['url'] if self._thumbnail else None

    @property
    def image(self) -> str | None:
        return self._image['url'] if self._image else None

    @property
    def footer(self) -> Dict[str,str] | None:
        return self._footer if self._footer else None

    def set_author(self, *, name: Optional[str] = None, icon_url: Optional[str] = None):
        self._author = {}
        if name:
            self._author["name"] = name
        if icon_url:
            self._author["icon_url"] = icon_url
        return self

    def set_thumbnail(self, *, url: str):
        self._thumbnail = {"url": url}
        return self

    def set_image(self, *, url: str):
        self._image = {"url": url}
        return self

    def set_footer(self, *, text: Optional[str] = None, icon_url: Optional[str] = None):
        self._footer = {}
        if text:
            self._footer["text"] = text
        if icon_url:
            self._footer["icon_url"] = icon_url
        return self

    def to_dict(self) -> Dict:
        data = {
            "title": self.title,
            "description": self.description,
            "url": self.url,
            "color": self.color,
            "author": self._author,
            "thumbnail": self._thumbnail,
            "image": self._image,
            "footer": self._footer,
        }
        return {k: v for k, v in data.items() if v is not None}

    @classmethod
    def from_dict(cls, data: dict):
        self = cls.__new__(cls)

        self.title = data.get('title')
        self.description = data.get('description')
        self.url = data.get('url')
        self._author = data.get('author')
        self._thumbnail = data.get('thumbnail')
        self._image = data.get('image')
        self._footer = data.get('footer')



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from __future__ import annotations

import asyncio
import threading
import aiohttp
import time

from typing import TYPE_CHECKING, Any, Optional, Dict, Callable
from loguru import logger

from teamly import utils

from .state import ConnectionState

if TYPE_CHECKING:
    from .client import Client


class GatewayRatelimiter:
    def __init__(self, count: int = 110, per: float = 60.0) -> None:
        # The default is 110 to give room for at least 10 heartbeats per minute
        self.max: int = count
        self.remaining: int = count
        self.window: float = 0.0
        self.per: float = per
        self.lock: asyncio.Lock = asyncio.Lock()

    def is_ratelimited(self) -> bool:
        current = time.time()
        if current > self.window + self.per:
            return False
        return self.remaining == 0

    def get_delay(self) -> float:
        current = time.time()

        if current > self.window + self.per:
            self.remaining = self.max

        if self.remaining == self.max:
            self.window = current

        if self.remaining == 0:
            return self.per - (current - self.window)

        self.remaining -= 1
        return 0.0

    async def block(self) -> None:
        async with self.lock:
            delta = self.get_delay()
            if delta:
                logger.warning('WebSocket is ratelimited, waiting {} seconds',delta)
                await asyncio.sleep(delta)


class KeepAliveHandler(threading.Thread):

    def __init__(self, ws: TeamlyWebSocket,interval: Optional[float] = None):
            super().__init__(daemon=True)
            self.ws: TeamlyWebSocket = ws
            self.interval: Optional[float] = interval
            self.behind_msg: str = 'Can\'t keep up, websocket is {} behind.'
            self._stop_event: threading.Event = threading.Event()
            self._last_send: float = time.perf_counter()
            self._last_ack: float = time.perf_counter()
            self.latency: float = float('inf')

    def run(self):
        while not self._stop_event.wait(self.interval):
            if time.perf_counter() - self._last_ack > self.interval * 2: #type: ignore
                logger.warning("[KeepAliveHandler] ACK not received, connection may be dead.")
                continue

            self._last_send = time.perf_counter()

            data = self.get_payload()
            coro = self.ws.send_heartbeat(data)
            future = asyncio.run_coroutine_threadsafe(coro, self.ws.loop)
            try:
                future.result(timeout=10)
            except Exception as e:
                logger.error("[KeepAliveHandler] send error:", e)

    def get_payload(self) -> Dict[str, Any]:
        return {
            "t": "HEARTBEAT",
            "d": {}
        }

    def stop(self):
        self._stop_event.set()

    def ack(self):
        ack_time = time.perf_counter()
        self._last_ack = ack_time
        self.latency = ack_time - self._last_send
        if self.latency > 10:
            logger.warning(self.behind_msg,self.latency)

    def tick(self):
        pass

class TeamlyWebSocket:

    if TYPE_CHECKING:
        _connection: ConnectionState
        _dispatch_parsers: Dict[str, Callable[..., Any]]

    def __init__(self, socket: aiohttp.ClientWebSocketResponse,*, loop: asyncio.AbstractEventLoop) -> None: #type: ignore
        self.socket: aiohttp.ClientWebSocketResponse = socket
        self._keep_alive: Optional[KeepAliveHandler] = None
        self.loop: asyncio.AbstractEventLoop = loop
        self._close_code: Optional[int] = None

    @classmethod
    async def from_client(cls, client: Client):
        socket = await client.http.ws_connect()
        ws = cls(socket, loop=client.loop)

        ws._connection = client._connection
        ws._dispatch_parsers = client._connection.parsers

        await ws.poll_event()

        return ws

    async def poll_event(self):
        '''
            Continuously polls the WebSocket connection for incoming messages.

            Depending on the message type (text, binary, error, or close),
            it delegates the message to the appropriate handler or logs it.

            This function is meant to be run in an event loop (e.g., inside a task).
        '''

        try:
            msg = await self.socket.receive()
            if msg.type is aiohttp.WSMsgType.TEXT:
                await self.received_message(msg.data)
            elif msg.type is aiohttp.WSMsgType.BINARY:
                await self.received_message(msg.data)
            elif msg.type is aiohttp.WSMsgType.ERROR:
                logger.error("Received: ", msg)
            elif msg.type in (
                aiohttp.WSMsgType.CLOSE,
                aiohttp.WSMsgType.CLOSED,
                aiohttp.WSMsgType.CLOSING
            ):
                logger.debug("Received: {}",msg)
        except Exception as e:
            logger.error("Exception error {}",e)

    async def received_message(self, msg: Any):
        if type(msg) is bytes:
            msg = msg.decode('utf-8')

        if msg is None:
            return

        msg = utils._to_json(msg)

        event = msg["t"]
        data = msg["d"]

        # if self._keep_alive:
        #     self._keep_alive.tick()

        if event == "HEARTBEAT_ACK":
            if self._keep_alive:
                self._keep_alive.ack()
            return

        if event == "READY":
            interval = data["heartbeatIntervalMs"] / 1000
            self._keep_alive = KeepAliveHandler(ws=self,interval=interval)
            await self.socket.send_json(self._keep_alive.get_payload())
            self._keep_alive.start()

        try:
            func = self._dispatch_parsers[event]
        except KeyError:
            logger.debug("Unknown event {}",event)
        else:
            func(data)

        #print(json.dumps(msg,indent=4,ensure_ascii=False))

    async def close(self, code: int = 4000) -> None:
        if self._keep_alive:
            self._keep_alive.stop()
            self._keep_alive = None

        self._close_code = code
        await self.socket.close(code=code)

    async def send_heartbeat(self, data: Any):
        await self.socket.send_str(data)



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import re
import random
from typing import Tuple

__all__ = ['Color']

class Color:
    """
    Represents a 24-bit RGB color.

    The color is internally stored as an integer between 0x000000 and 0xFFFFFF.
    This class provides utility methods for converting between integer, hex,
    RGB tuples, and string representations.

    Example:
        c = Color(0xFF0000)
        print(c.to_rgb())       # (255, 0, 0)
        print(str(c))           # <Color #FF0000>
    """

    __slots__ = ('value')

    def __init__(self, value: int):
        """
        Initializes a Color instance from an integer.

        Parameters:
            value (int): Integer between 0x000000 and 0xFFFFFF.

        Raises:
            ValueError: If the value is out of range.
        """

        if not (0 <= value <= 0xFFFFFF):
            raise ValueError("Color value must be between 0x000000 and 0xFFFFFF")
        self.value = value

    def __int__(self):
        """
        Returns the integer value of the color.

        Returns:
            int: The 24-bit RGB integer.
        """

        return self.value

    def __repr__(self):
        """
        Returns the string representation of the color in hex format.

        Returns:
            str: A string like <Color #RRGGBB>.
        """

        return f"<Color #{self.value:06X}>"

    def to_rgb(self) -> Tuple[int, int, int]:
        """
        Converts the color to an (R, G, B) tuple.

        Returns:
            tuple: A tuple of 3 integers between 0 and 255.
        """

        r = (self.value >> 16) & 0xFF
        g = (self.value >> 8) & 0xFF
        b = self.value & 0xFF
        return (r, g, b)

    @classmethod
    def from_rgb(cls, r: int, g: int, b: int) -> "Color":
        """
        Creates a Color instance from RGB values.

        Parameters:
            r (int): Red component (0–255).
            g (int): Green component (0–255).
            b (int): Blue component (0–255).

        Returns:
            Color: A new Color instance.

        Raises:
            ValueError: If any value is out of the 0–255 range.
        """

        if not all(0 <= n <= 255 for n in (r, g, b)):
            raise ValueError("RGB values must be between 0 and 255")
        return cls((r << 16) + (g << 8) + b)

    @classmethod
    def from_hex(cls, hex_str: str) -> "Color":
        """
        Creates a Color from a hex string.

        Supports 3-digit (e.g., "#f00") or 6-digit (e.g., "#ff0000") formats.

        Parameters:
            hex_str (str): Hex color string (with or without '#').

        Returns:
            Color: A new Color instance.

        Raises:
            ValueError: If the string is not a valid 3- or 6-digit hex code.
        """

        hex_str = hex_str.strip().lstrip("#")
        if len(hex_str) not in (3, 6):
            raise ValueError("Hex string must be 3 or 6 characters")
        if len(hex_str) == 3:
            hex_str = "".join(c * 2 for c in hex_str)
        return cls(int(hex_str, 16))

    @classmethod
    def from_str(cls, value: str) -> "Color":
        """
        Parses a color from a string.

        Supported formats:
            - "#RRGGBB"
            - "#RGB"
            - "rgb(R, G, B)"

        Parameters:
            value (str): A color string in hex or RGB format.

        Returns:
            Color: A new Color instance.

        Raises:
            ValueError: If the string format is invalid or unrecognized.
        """
        value = value.strip()
        if value.startswith("#"):
            return cls.from_hex(value)
        if value.lower().startswith("rgb"):
            match = re.fullmatch(r"rgb\(\s*(\d{1,3})\s*,\s*(\d{1,3})\s*,\s*(\d{1,3})\s*\)", value)
            if not match:
                raise ValueError("Invalid RGB string format")
            r, g, b = map(int, match.groups())
            return cls.from_rgb(r, g, b)
        raise ValueError(f"Unknown color format: {value}")

    @classmethod
    def red(cls) -> "Color":
        """Returns a standard red color (#E74C3C)."""
        return cls(0xE74C3C)

    @classmethod
    def green(cls) -> "Color":
        """Returns a standard green color (#2ECC71)."""
        return cls(0x2ECC71)

    @classmethod
    def blue(cls) -> "Color":
        """Returns a standard blue color (#3498DB)."""
        return cls(0x3498DB)

    @classmethod
    def yellow(cls) -> "Color":
        """Returns a standard yellow color (#F1C40F)."""
        return cls(0xF1C40F)

    @classmethod
    def orange(cls) -> "Color":
        """Returns a standard orange color (#E67E22)."""
        return cls(0xE67E22)

    @classmethod
    def purple(cls) -> "Color":
        """Returns a standard purple color (#9B59B6)."""
        return cls(0x9B59B6)

    @classmethod
    def teal(cls) -> "Color":
        """Returns a standard teal color (#1ABC9C)."""
        return cls(0x1ABC9C)

    @classmethod
    def pink(cls) -> "Color":
        """Returns a standard pink color (#FF69B4)."""
        return cls(0xFF69B4)

    @classmethod
    def random(cls) -> "Color":
        """
        Returns a randomly generated Color.

        Returns:
            Color: A color with a random 24-bit RGB value.
        """

        return cls(random.randint(0x000000, 0xFFFFFF))








from typing import Dict, Optional, TypedDict


class Role(TypedDict):
    id: str
    teamId: str
    name: str
    iconUrl: Optional[str]
    color: str
    color2: Optional[str]
    permissions: int
    priority: int
    createdAt: str
    updatedAt: Optional[str]
    isDisplayedSeparately: bool
    isSelfAssignable: bool
    iconEmojiId: Optional[str]
    mentionable: bool
    botScope: Optional[Dict[str,str]]



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from __future__ import annotations

from loguru import logger




from .types.team import TeamPayload, TeamGames as TeamGamesPayload
from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from .state import ConnectionState
    from .role import Role


__all__ = ['Team']

class TeamGames:

    __slots__ = (
        'id',
        'platforms',
        'region'
    )

    def __init__(self, data: TeamGamesPayload) -> None:
        self.id: str = data['id']
        self.platforms: List[str] = data['platforms']
        self.region: str = data['region']

class Team:

    __slots__ = (
        '_state',
        'id',
        'name',
        '_profile_picture',
        'banner',
        'description',
        '_is_verified',
        '_is_safe_for_teen',
        '_is_suspended',
        '_created_by',
        '_default_channel_id',
        'games',
        '_is_discoverable',
        '_is_tournament',
        '_discoverable_invite',
        '_created_at',
        '_member_count'
    )

    def __init__(self,*, state: ConnectionState, data: TeamPayload) -> None:
        self._state = state
        self._update(data)

    def _update(self, data: TeamPayload):
        self.id: str = data['id']
        self.name: str = data['name']
        self._profile_picture: Optional[str] = data.get('profilePicture')
        self.banner: Optional[str] = data.get('banner')
        self.description: Optional[str] = data.get('description')

        self._is_verified: bool = data['isVerified']
        self._is_safe_for_teen: bool = data.get('isSafeForTeen', False)
        self._is_suspended: bool = data['isSuspended']
        self._created_by: str = data['createdBy']
        self._default_channel_id: str = data.get('defaultChannelId')
        self.games: List[TeamGames] = [TeamGames(g) for g in data.get('games')]
        self._is_discoverable: bool = data.get('idDiscoverable', False)
        self._is_tournament: bool = data.get('isTournament', False)
        self._discoverable_invite: Optional[str] = data.get('discoverableInvite')
        self._created_at: str = data['createdAt']
        self._member_count: int = data['memberCount']

    #Team

    async def edit(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        banner: Optional[str] = None,
        profilePicture: Optional[str] = None
    ):
        if len(name) > 12:
            raise ValueError("\'name\' must be smaller or equel then 12 characters")

        if len(description) > 1000:
            raise ValueError("\'description\' must be smaller or equel then 1000 characters")

        payload = {
            k:v
            for k,v in locals().items()
            if v is not None
        }

        try:
            await self._state.http.update_team(self.id, payload=payload)
        except Exception as e:
            logger.error(f"Exception error: {e}")


    def info(self) -> str:
        return (
            f"""
            Team:
                ID: {self.id}
                Name: {self.name}
                ProfilePicture: {self.profile_picture[:10] + '...' if self.profile_picture else 'N/A'}
                Banner: {self.banner[:10] + '...' if self.banner else 'N/A'}
                Description: {self.description[:25] + '...' if self.description else 'N/A'}
                IsVerified: {self.is_verified}
                IsSafeForTeen: {self._is_safe_for_teen}
                IsSuspended: {self.is_suspended}
                CreatedBy: {self._created_by}
                CreatedAt: {self._created_at}
                DefaultChannel: {self._default_channel_id if self._default_channel_id else 'N/A'}
                Games: {[t.id for t in self.games]}
                IsDiscoverable: {self._is_discoverable}
                IsTournament: {self._is_tournament}
                DiscoverableInvite: {self._discoverable_invite}
            """
        )

    #Member

    def fetch_members(self):
        return self._state.cache.get_members(teamId=self.id)

    def get_members_count(self):
        return len(self._state.cache.get_members(teamId=self.id))

    def get_member(self, userId: str):
        return self._state.cache.get_member(teamId=self.id, userId=userId)




    async def ban(self, userId: str, reason: str):
        return await self._state.http.ban(teamId=self.id, userId=userId, reason=reason)

    async def unban(self, userId: str):
        return await self._state.http.unban(teamId=self.id, userId=userId)

    async def get_banned_users(self, teamId: str, limit: int = 10):
        return await self._state.http.get_banned_users(teamId=teamId, limit=limit)


    async def kick(self, userId: str):
        return await self._state.http.kick_member(teamId=self.id, userId=userId)


    #Role

    async def add_role(self, role: Role):
        return await self._state.http.create_role(teamId=self.id, payload=role.to_dict())

    async def remove_role(self, roleId: str):
        return await self._state.http.delete_role(teamId=self.id, roleId=roleId)

    async def list_roles(self):
        return await self._state.http.get_roles(teamId=self.id)

    async def assigne_role(self, userId: str, roleId: str):
        return await self._state.http.add_role_to_member(teamId=self.id, userId=userId, roleId=roleId)

    async def unassigne_role(self, userId: str, roleId: str):
        return await self._state.http.remove_role_from_member(teamId=self.id, userId=userId, roleId=roleId)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return (
            f"<Team id={self.id} name={self.name} description={self.description}"
            f" isVerified={self.is_verified} isDiscoverable={self.is_discoverable}>"
        )



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from __future__ import annotations

import asyncio
from collections import OrderedDict
import json
from typing import Dict, Optional, Union, Any, TYPE_CHECKING

from loguru import logger

from .channel import TextChannel, VoiceChannel, _channel_factory
from .http import HTTPClient
from .message import Message
from .team import Team
from .member import Member
from .user import ClientUser

if TYPE_CHECKING:
    from .state import ConnectionState
    from .channel import (
        TextChannel,
        VoiceChannel
    )
    from .message import Message
    MessageAbleChannel = Union[TextChannel]
    Channel = Union[TextChannel, VoiceChannel]

__all__ = [
    "Cache"
]

class Cache:

    def __init__(self, state: ConnectionState) -> None:
        self._state: ConnectionState = state
        self._http: HTTPClient = self._state.http
        self.maxLength: int = 100
        self.clear()

    def clear(self):
        self.__user: Optional[ClientUser] = None
        self.__teams: Dict[str, Team] = {}
        self.__channels: Dict[str, Dict[str, Channel]] = {}
        self.__messages: Dict[str, Dict[str, OrderedDict[str, Message]]] = {}
        self.__members: Dict[str, Dict[str, Member]] = {}

    async def setup_cache(self, data: Any):
        #get ClietnUser Payload
        self.__user = ClientUser(state=self._state, data=data['user'])
        logger.info(f"Bot connected as {self.__user.username!r}")
        await self.__fetch_teams(data['teams'])
        tasks = []
        for team in self.__teams:
            tasks.append(self.__fetch_channels(team))
            tasks.append(self.__fetch_team_members(team))

        await asyncio.gather(*tasks)

    async def __fetch_teams(self, teams: Dict[str, Any]):
        for team in teams:
            self.__teams[team['id']] = Team(state=self._state, data=team)

    async def __fetch_channels(self, teamId: str):
        channels = await self._http.get_channels(teamId)
        channels = json.loads(channels)

        for data in channels['channels']:
            factory = _channel_factory(data['type'])
            channel: Channel = None
            team: Team = None

            if teamId not in self.__channels:
                self.__channels[teamId] = {}

            if factory:
                team = self.__teams[teamId]
                self.__channels[teamId][data['id']] = channel = factory(state=self._state,team=team, data=data)

            if teamId not in self.__messages:
                self.__messages[teamId] = OrderedDict()

            if data['type'] == 'text':
                self.__messages[teamId][channel.id] = await self.__fetch_channel_messages(channel=channel)

    async def __fetch_channel_messages(self, channel: MessageAbleChannel):
        try:
            messages = await self._http.get_channel_messages(channelId=channel.id, limit=50)
            messages = json.loads(messages)

            message_dict = {}

            if messages['messages']:
                for message in messages['messages']:
                    message_dict[message['id']] = Message(state=self._state, channel=channel, data=message)

            if messages['replyMessages']:
                for message in messages['replyMessages']:
                    message_dict[message['id']] = Message(state=self._state, channel=channel, data=message)

            return message_dict
        except Exception as e:
            logger.error(f"Exception error: {e}")
            return {}

    async def __fetch_team_members(self, teamId: str):
        members = await self._http.get_member(teamId, "")
        members = json.loads(members)

        if teamId not in self.__members:
            self.__members[teamId] = {}

        for member in members['members']:
            if member['id'] not in self.__members:
                self.__members[member['id']] = Member(state=self._state, data=member)


    #Team Cache
    def get_team(self, teamId):
        if teamId in self.__teams:
            return self.__teams[teamId]


    #Channel Cache
    def add_channel(self, teamId: str, channelId: str, channel: MessageAbleChannel):
        if teamId not in self.__channels:
            self.__channels[teamId] = {}

        if channelId not in self.__channels[teamId]:
            self.__channels[teamId][channelId] = channel
            logger.opt(colors=True).debug(f"<cyan>Added channel {channelId!r} to cache successfuly</cyan>")

    def delete_channel(self, teamId: str, channelId: str):
        if channelId in self.__channels[teamId]:
            self.__channels[teamId].pop(channelId)
            logger.opt(colors=True).debug(f"<cyan>Deleted channel {channelId!r} from cache successfuly</cyan>")

    def update_channel(self, teamId: str, channelId: str, channel: MessageAbleChannel):
        if channelId in self.__channels[teamId]:
            self.__channels[teamId][channelId] = channel
            logger.opt(colors=True).debug(f"<cyan>Updated channel {channelId!r} from cache successfuly</cyan>")

    def get_channel(self, teamId: str, channelId: str) -> Channel | None:
        try:
            if channelId in self.__channels[teamId]:
                return self.__channels[teamId][channelId]
        except Exception as e:
            logger.error(f"Exception error: {e}")

    #Voice Channel
    def voice_participants_joined(self,teamId: str, channelId: str, participantId: str):
        if self.__channels[teamId][channelId]:
            voice = self.__channels[teamId][channelId]
            if not any(p.get('id') == participantId for p in voice._participants):
                voice._participants.append({"id": participantId})

    def voice_participants_leaved(self, teamId:str, channelId: str, participantId: str):
        if self.__channels[teamId][channelId]:
            voice = self.__channels[teamId][channelId]
            for par in voice._participants:
                if par.get('id') == participantId:
                    voice._participants.remove({"id": participantId})


    #Message Cache

    def add_message(self, teamId: str, channelId: str, message: Message):
        if channelId in self.__messages[teamId]:
            self.__messages[teamId][channelId][message.id] = message

        if len(self.__messages[teamId][channelId]) > self.maxLength:
            self.__messages[teamId][channelId].popitem(last=False)

    def update_message(self, teamId: str, channelId: str, message: Message):
        if channelId in self.__messages[teamId]:
            if message.id in self.__messages[teamId][channelId]:
                upt_message = self.__messages[teamId][channelId][message.id]
                self.__messages[teamId][channelId][message.id] = message
                logger.opt(colors=True).debug(f"<cyan>updated channel message {message.id!r} from cache successfuly</cyan>")
                return upt_message


    def delete_message(self, teamId: str, channelId: str, messageId: str):
        if channelId in self.__messages[teamId]:
            if messageId in self.__messages[channelId][messageId]:
                message = self.__messages[teamId][channelId].pop(messageId)
                logger.opt(colors=True).debug(f"<cyan>deleted channel message {messageId!r} from cache successfuly</cyan>")
                return message



    #Member

    def get_members(self, teamId: str):
        return self.__members[teamId].values

    def get_member(self, teamId: str, userId: str):
        return self.__members[teamId][userId] if userId in self.__members[teamId] else None

    def add_member(self, teamId: str, member: Member):
        if member.id not in self.__members[teamId]:
            self.__members[teamId][member.id] = member

    def delete_member(self, teamId: str, memberId: str):
        if memberId in self.__members[teamId]:
            self.__members[teamId].pop(memberId)



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from __future__ import annotations

import teamly.abc
from teamly.todo import TodoItem

from .enums import ChannelType

from .types.channel import (
    TextChannel as TextChannelPayload,
    DMChannel as DMChannelPayload,
    VoiceChannel as VoiceChannelPayload,
    AnnouncementChannel as AnnouncementChannelPayload,
    TodoChannel as TodoChannelPayload
)
from typing import TYPE_CHECKING, Dict, Any, List, Optional

if TYPE_CHECKING:
    from .state import ConnectionState
    from .team import Team


__all__ = [
    'TextChannel',
    'DMChannel',
    'VoiceChannel',
    'AnnouncementChannel',
    'TodoChannel'
]

class TextChannel(teamly.abc.MessageAble):

    __slots__ = (
        '_state',
        'team',
        'id',
        'type',
        'name',
        'description',
        'created_by',
        'parent_id',
        'priority',
        'rate_limit_per_user',
        'created_at',
        'permissions',
        'additional_data'
    )

    def __init__(
        self,
        *,
        state: ConnectionState,
        team: Team,
        data: TextChannelPayload
    ) -> None:
        self._state: ConnectionState = state
        self.team: Team = team
        self._update(data)

    def _update(self, data: TextChannelPayload):
        self.id: str = data['id']
        self.type: str = data['type']
        self.name: str = data['name']

        self.description: Optional[str] = data.get('description', None)
        self.created_by: str = data.get('createdBy')
        self.parent_id: Optional[str] = data.get('parentId', None)
        self.priority: int = data['priority']
        self.rate_limit_per_user: int = data['rateLimitPerUser']
        self.created_at: str = data['createdAt']
        self.permissions: Dict[str,Any] = data['permissions'].get('role', {})
        self.additional_data: Dict[str,Any] = data.get('additionalData', {})


    def __repr__(self) -> str:
        return f"<TextChannel id={self.id} name={self.name!r} type={self.type} teamId={self.team.id}>"

    async def delete_message(self, messageId: str):
        return await self._state.http.delete_message(self.id, messageId)

    async def react_message(self, messageId: str, emojiId: str):
        return await self._state.http.react_to_message(self.id, messageId, emojiId)


class DMChannel:

    def __init__(self, state: ConnectionState, data: DMChannelPayload) -> None:
        self._state: ConnectionState = state
        

class VoiceChannel:

    __slots__ = (
        '_state',
        'team',
        'id',
        'type',
        'name',
        'description',
        'created_by',
        'parent_id',
        'priority',
        '_participants',
        'created_at',
        'permissions',
        'additional_data'
    )

    def __init__(
        self,
        *,
        state: ConnectionState,
        team: Team,
        data: VoiceChannelPayload
    ) -> None:
        self._state: ConnectionState = state
        self.team: Team = team
        self._update(data)

    def _update(self, data: VoiceChannelPayload):
        self.id: str = data['id']
        self.type: str = data['type']
        self.name: str = data['name']

        self.description: Optional[str] = data.get('description', None)
        self.created_by: str = data.get('createdBy')
        self.parent_id: Optional[str] = data.get('parentId', None)
        self.priority: int = data['priority']
        self._participants: List[Dict[str,str]] = data.get('participants', [])
        self.created_at: str = data['createdAt']
        self.permissions: Dict[str,Any] = data['permissions'].get('role', {})
        self.additional_data: Dict[str,Any] = data.get('additionalData', {})

    @property
    def participants(self):
        return [p.get('id') for p in self._participants if self._participants] if self._participants else []


    def __repr__(self) -> str:
        return f"<VoiceChannel id={self.id} name={self.name!r} type={self.type} teamId={self.team.id}>"


class AnnouncementChannel:

    __slots__ = (
        '_state',
        'team',
        'id',
        'type',
        'name',
        'description',
        'created_by',
        'parent_id',
        'priority',
        'rate_limit_per_user',
        'created_at',
        'permissions',
        'additional_data'
    )

    def __init__(
        self,
        *,
        state: ConnectionState,
        team: Team,
        data: AnnouncementChannelPayload
    ) -> None:
        self._state: ConnectionState = state
        self.team: Team = team
        self._update(data)

    def _update(self, data: AnnouncementChannelPayload):
        self.id: str = data['id']
        self.type: str = data['type']
        self.name: str = data['name']

        self.description: Optional[str] = data.get('description', None)
        self.created_by: str = data.get('createdBy')
        self.parent_id: Optional[str] = data.get('parentId', None)
        self.priority: int = data['priority']
        self.created_at: str = data['createdAt']
        self.permissions: Dict[str,Any] = data['permissions'].get('role', {})
        self.additional_data: Dict[str,Any] = data.get('additionalData', {})

    def __repr__(self) -> str:
        return f"<VoiceChannel id={self.id} name={self.name!r} type={self.type} teamId={self.team.id}>"

class TodoChannel:
    def __init__(
        self,
        *,
        state: ConnectionState,
        team: Team,
        data: TodoChannelPayload
    ) -> None:
        self._state: ConnectionState = state
        self.team: Team = team
        self._update(data)

    def _update(self, data: TodoChannelPayload):
        self.id: str = data['id']
        self.type: str = data['type']
        self.name: str = data['name']

        self.description: Optional[str] = data.get('description', None)
        self.created_by: str = data.get('createdBy')
        self.parent_id: Optional[str] = data.get('parentId', None)
        self.priority: int = data['priority']
        self.created_at: str = data['createdAt']
        self.permissions: Dict[str,Any] = data['permissions'].get('role', {})
        self.additional_data: Dict[str,Any] = data.get('additionalData', {})

    async def fetch_todoitems(self):
        return self._state.http.get_todo_items(channelId=self.id)

    async def create_todo(self, content: str):
        if len(content) >= 256:
            raise ValueError("Content is too long, max 256 characters")

        todo = await self._state.http.create_todo_item(channelId=self.id, content=content)
        return TodoItem(state=self._state, channel=self, data=todo['todo'])

    async def delete_todo(self, todoId: str):
        return await self._state.http.delete_todo_item(channelId=self.id, todoId=todoId)

    async def clone_todoitem(self, todoId: str):
        clone = await self._state.http.clone_todo_item(channelId=self.id, todoId=todoId)
        return TodoItem(state=self._state,channel=self,data=clone['todo'])


    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<TodoChannel id={self.id} name={self.name!r} type={self.type} teamId={self.team.id}>"


class WatchStreamChannel:
    def __init__(
        self,
        *,
        state: ConnectionState,
        team: Team,
        data: TodoChannelPayload
    ) -> None:
        self._state: ConnectionState = state
        self.team: Team = team
        self._update(data)

    def _update(self, data: TodoChannelPayload):
        self.id: str = data['id']
        self.type: str = data['type']
        self.name: str = data['name']

        self.description: Optional[str] = data.get('description', None)
        self.created_by: str = data.get('createdBy')
        self.parent_id: Optional[str] = data.get('parentId', None)
        self.priority: int = data['priority']
        self.created_at: str = data['createdAt']
        self.permissions: Dict[str,Any] = data['permissions'].get('role', {})
        self.additional_data: Dict[str,Any] = data.get('additionalData', {})

    def __repr__(self) -> str:
        return f"<WatchStreamChannel id={self.id} name={self.name!r} type={self.type} teamId={self.team.id}>"


@staticmethod
def _channel_factory(type: str) -> Any | None:
    if ChannelType.TEXT == type:
        return TextChannel
    elif ChannelType.VOICE == type:
        return VoiceChannel
    elif ChannelType.ANNOUNCEMENT == type:
        return AnnouncementChannel
    elif ChannelType.TODO == type:
        return TodoChannel
    elif ChannelType.WATCHSTREAM == type:
        return WatchStreamChannel
    else:
        return None



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from __future__ import annotations

import teamly.abc

from .reaction import PartialReaction
from .user import User
from .embed import Embed
from .types.message import Message as MessagePayload
from typing import TYPE_CHECKING, Dict, List, Optional, Union

if TYPE_CHECKING:
    from teamly.state import ConnectionState

    from .channel import TextChannel

    MessageAbleChannel = Union[TextChannel]


class Message(teamly.abc.MessageAble):

    __slots__ = (
        '_state',
        'id',
        'channel',
        'type',
        'content',
        '_attachment',
        '_created_by',
        'edited_at',
        '_reply_to',
        '_embeds',
        '_emojis',
        '_reactions',
        'nonce',
        'created_at',
        '_mentions'
    )

    def __init__(
        self,
        state: ConnectionState,
        *,
        channel: MessageAbleChannel,
        data: MessagePayload
    ) -> None:
        super().__init__(state=state)
        self._state: ConnectionState = state
        self.channel: MessageAbleChannel = channel
        self.from_dict(data)

    def from_dict(self, data: MessagePayload):
        self.id: str = data['id']
        self.type: str = data['type']
        self.content: Optional[str] = data.get('content')

        self._attachment: Optional[List[Dict[str,str]]] = data.get('attachments')
        self._created_by: User = User(state=self._state, data=data['createdBy'])
        self.edited_at: Optional[str] = data.get('editedAt')
        self._reply_to: Optional[str] = data.get('replyTo')
        self._embeds: Optional[List[Embed]] = data.get('embeds')
        self._emojis: Optional[List[Dict[str,str]]] = data.get('emojis')
        self._reactions: Optional[List[Dict[str,str]]] = data.get('reactions')
        self.nonce: Optional[str] = data.get('nonce')
        self.created_at: str = data['createdAt']
        self._mentions: Dict[str,List[str]] = data.get('mentions')

    @property
    def author(self) -> User:
        return self._created_by

    @property
    def attachment(self):
        if self._attachment:
            return [x.get('url') for x in self._attachment]
        else:
            return []

    @property
    def repliedTo(self):
        if self._reply_to:
            return self._reply_to
        else:
            return None

    @property
    def embeds(self):
        if self._embeds:
            return [Embed.from_dict(e) for e in self._embeds]
        else:
            return []

    @property
    def emojis(self):
        if self._emojis:
            return [e.get('emojiId') for e in self._emojis]
        else:
            return []

    @property
    def reactions(self):
        if self._reactions:
            return [PartialReaction(r) for r in self._reactions]
        else:
            return []

    @property
    def mentions(self):
        if self._mentions:
            return self._mentions['users']

    def to_dict(self):
        result = {
            "id": self.id,
            "channelId": self.channel.id,
            "type": self.type,
            "content": self.content,
            "createdBy": self.author.to_dict(),
            "createdAt": self.created_at
        }
        if self.attachment:
            result['attachment'] = self._attachment
        if self.edited_at:
            result['editedAt'] = self.edited_at
        if self._reply_to:
            result['replyTo'] = self._reply_to
        if self.embeds:
            result['embeds'] = self.embeds
        if self.emojis:
            result['emojis'] = self.emojis
        if self.reactions:
            result['reactions'] = self.reactions
        if self.nonce:
            result['nonce'] = self.nonce
        if self.mentions:
            result['mentions'] = self.mentions

        return result

    def __repr__(self) -> str:
        return (
            f"<Message id={self.id} channelId={self.channel.id} type={self.type} content={self.content} "
            f"createdBy={self.author.username}>"
        )

    class PrivateMessage:
        pass



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from __future__ import annotations

import asyncio
import aiohttp
import json
import time






from .utils import MISSING
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Union, Optional
from urllib.parse import quote
from loguru import logger

if TYPE_CHECKING:
    from .embed import Embed


async def message_handler(
    content: Optional[str],
    *,
    embeds: Optional[List[Embed]],
    attachment: List[Dict[str,Any]],
    replyTo: str
):
    payload = {}

    if len(content) > 2000:
        raise ValueError("the content must equel or lower then 2000 characters")
    else:
        payload['content'] = content

    if embeds:
       payload['embeds'] = embeds

    if attachment:
       payload['attachment'] = attachment

    if replyTo:
       payload['replyTo'] = replyTo

    return payload


async def json_or_text(response: aiohttp.ClientResponse) -> Union[Dict[str, Any], str]:
    text = await response.text(encoding='utf-8')
    try:
        if response.headers['content-type'] == 'application/json':
            #logger.warning("returning as json file")
            return json.loads(text)
    except KeyError:
        pass

    #logger.warning("returning as text file")
    return text

class Route:
    '''
        Represents an API route with a method and path, used to build full request URLs
        for the Teamly API.

        Example:
            Route("GET", "/channels/{channel_id}", channel_id="1234")
            → https://api.teamly.one/api/v1/channels/1234

        Attributes:
            BASE_URL (str): The base URL for all API requests.
            method (str): The HTTP method (e.g., "GET", "POST").
            path (str): The API path, possibly containing placeholders.
            url (str): The fully constructed request URL.
    '''

    BASE_URL = "https://api.teamly.one/api/v1"

    def __init__(self, method:str, path: str, **params: Any) -> None:
        self.method = method
        self.path = path

        url = self.BASE_URL + self.path
        if params:
            url = url.format_map({
                k: quote(v, safe='') if isinstance(v, str) else v
                for k, v in params.items()
            })
        self.url: str = url

class RateLimit:
    """Simple rate limiter for HTTP requests."""

    def __init__(self, count: int = 5, per: float = 1.0) -> None:
        self.max: int = count
        self.remaining: int = count
        self.window: float = 0.0
        self.per: float = per
        self.lock: asyncio.Lock = asyncio.Lock()

    def _get_delay(self) -> float:
            current = time.time()

            if current > self.window + self.per:
                self.remaining = self.max

            if self.remaining == self.max:
                self.window = current

            if self.remaining == 0:
                return self.per - (current - self.window)

            self.remaining -= 1
            return 0.0

    async def block(self) -> None:
        async with self.lock:
            delay = self._get_delay()
            if delay:
                logger.warning('HTTP client is ratelimited, waiting {} seconds', delay)
                await asyncio.sleep(delay)


class HTTPClient:

    def __init__(self,loop: asyncio.AbstractEventLoop, *, rate_limit: int = 5, per: float = 1.0) -> None:
        self._session: aiohttp.ClientSession = MISSING
        self.token = None
        self.loop: asyncio.AbstractEventLoop = loop
        self._ratelimiter = RateLimit(rate_limit, per)

    async def static_login(self, token: str):
        logger.debug("static logging...")

        self.token = token
        self._session = aiohttp.ClientSession()

        try:
            data = await self.get_loggedIn_user
        except Exception as e:
            logger.debug("Exception error: {}",e)
        else:
            return data

    async def close(self):
        logger.debug("closing client session...")
        await self._session.close()

    async def ws_connect(self) -> aiohttp.ClientWebSocketResponse:
        logger.debug("creating ws connect...")

        kwargs = {
            "timeout": 30,
            "max_msg_size": 0,
            "headers": {
                "Authorization": f'Bot {self.token}'
            }
        }

        return await self._session.ws_connect(url="wss://api.teamly.one/api/v1/ws", **kwargs)

    async def request(self, route: Route, **kwargs) -> Any:
        method = route.method
        url = route.url

        await self._ratelimiter.block()

        #creating headers
        headers = {}

        if self.token is not None:
            headers["Authorization"] = f'Bot {self.token}'

        if 'json' in kwargs:
            headers["Content-Type"] = "application/json"
            kwargs['data'] = json.dumps(kwargs.pop('json'))

        kwargs["headers"] = headers

        data: Optional[Union[Dict[str,Any], str]] = None
        try:
            async with self._session.request(method, url, **kwargs) as response:
                logger.debug("Sending request {!r} {} with {}", method, url, kwargs)

                if response.status == 429:
                    retry_after = response.headers.get("Retry-After")
                    if retry_after is None:
                        payload = await json_or_text(response)
                        if isinstance(payload, dict) and "retry_after" in payload:
                            retry_after = payload.get("retry_after")
                    delay = float(retry_after or 1)
                    logger.warning('Rate limit hit, retrying after {} seconds', delay)
                    await asyncio.sleep(delay)
                    return await self.request(route, **kwargs)

                data = await json_or_text(response)

                status = response.status
                if 200 <= status < 300:
                    logger.debug("Request successful with status {}", status)
                elif 400 <= status < 500:
                        logger.warning("Client error with status {}", status)
                elif status >= 500:
                    logger.error("Server error with status {}", status)
                else:
                    logger.debug("Received status {}", status)

                return data
        except Exception as e:
            logger.error("Request failed: {}", e)
            raise


    #Core Resources

    #Channels
    async def get_channels(self, teamId: str):
        return await self.request(Route("GET","/teams/{teamId}/channels", teamId=teamId))

    async def create_channel(self, teamId: str, payload: Dict[str,Any]):
        r = Route("PUT", "/teams/{teamId}/channels", teamId=teamId)
        return await self.request(r,json=payload)

    async def delete_channel(self, teamId: str, channelId: str):
        r = Route("DELETE","/teams/{teamId}/channels/{channelId}", teamId=teamId, channelId=channelId)
        return await self.request(r)

    async def duplicate_channel(self, teamId: str, channelId: str):
        r = Route("POST","/teams/{teamId}/channels/{channelId}/clone", teamId=teamId, channelId=channelId)
        return await self.request(r)

    async def update_channel_priorities(self, teamId: str, channelId: str, payload: Dict[str, Any]):
        r = Route("PUT","/teams/{teamId}/channelspriority", teamId=teamId)
        return await self.request(r,json=payload)

    async def get_channel_by_Id(self, teamId: str, channelId: str):
        r = Route("GET","/teams/{teamId}/channels/{channelId}", teamId=teamId, channelId=channelId)
        return await self.request(r)




    async def update_channel(self, teamId: str, channelId: str, payload: Dict[str, Any]):
        r = Route("PATCH","/teams/{teamId}/channels/{channelId}", teamId=teamId, channelId=channelId)
        return await self.request(r, json=payload)

    async def update_channel_permissions(self, teamId: str, channelId: str, roleId: str, allow: int, deny: int):
        payload = {"allow": allow, "deny": deny}
        r = Route("POST","/teams/{teamId}/channels/{channelId}/permissions/role/{roleId}", teamId=teamId, channelId=channelId, roleId=roleId)
        return await self.request(r,json=payload)



    #Message
    async def create_message(self, channelId: str, payload: Dict[str,Any]):
        r = Route("POST","/channels/{channelId}/messages", channelId=channelId)
        return await self.request(r,json=payload)

    async def delete_message(self, channelId: str, messageId: str):
        r = Route("DELETE","/channels/{channelId}/messages/{messageId}",messageId=messageId, channelId=channelId)
        return await self.request(r)

    async def get_channel_messages(self, channelId: str, offset: str = '0', limit: str = '15'):
        r = Route("GET","/channels/{channelId}/messages" + f"?offset={offset}&limit={limit}", channelId=channelId)
        return await self.request(r)

    async def update_channel_message(self, channelId: str, messageId: str, payload: Dict[str, Any]):
        r = Route("PATCH","/channels/{channelId}/messages/{messageId}", channelId=channelId, messageId=messageId)
        return await self.request(r,json=payload)

    async def react_to_message(self, channelId: str, messageId: str, emojiId: str):
        r = Route("POST","/channels/{channelId}/messages/{messageId}/reactions/{emojiId}", channelId=channelId, messageId=messageId, emojiId=emojiId)
        return await self.request(r)

    async def delete_reaction_from_message(self, channelId: str, messageId: str, emojiId: str):
        r = Route("DELETE","/channels/{channelId}/messages/{messageId}/reactions/{emojiId}", channelId=channelId, messageId=messageId, emojiId=emojiId)
        return await self.request(r)

    async def get_channel_message_by_id(self, channelId: str, messageId: str):
        r = Route("GET","/channels/{channelId}/messages/{messageId}", channelId=channelId, messageId=messageId)
        return await self.request(r)


    ### Teams ###

    #Members
    async def add_role_to_member(self, teamId: str, userId: str, roleId: str):
        r = Route("POST","/teams/{teamId}/members/{userId}/roles/{roleId}", teamId=teamId, userId=userId, roleId=roleId)
        return await self.request(r)

    async def remove_role_from_member(self, teamId: str, userId: str, roleId: str):
        r = Route("DELETE","/teams/{teamId}/members/{userId}/roles/{roleId}", teamId=teamId, userId=userId, roleId=roleId)
        return await self.request(r)

    async def kick_member(self, teamId: str, userId: str):
        r = Route("DELETE","/teams/{teamId}/members/{userId}", teamId=teamId, userId=userId)
        return await self.request(r)

    async def get_member(self, teamId: str, userId: str):
        r = Route("GET","/teams/{teamId}/members/{userId}", teamId=teamId, userId=userId)
        return await self.request(r)

    #Bans
    async def get_banned_users(self, teamId: str, limit: str = 10) -> Dict[str, Any]:
        r = Route("GET","/teams/{teamId}/bans" + f"?limit={limit}", teamId=teamId)
        return await self.request(r)

    async def unban(self, teamId: str, userId: str) -> Dict[str, Any]:
        r = Route("DELETE","/teams/{teamId}/members/{userId}/ban", teamId=teamId, userId=userId)
        return await self.request(r)

    async def ban(self, teamId: str, userId: str, reason: str) -> Dict[str, Any]:
        payload = {"reason": reason}
        r = Route("POST","/teams/{teamId}/members/{userId}/ban", teamId=teamId, userId=userId)
        return await self.request(r, json=payload)



    async def get_team(self, teamId: str):
        r = Route("GET","/teams/{teamId}/details", teamId=teamId)
        return await self.request(r)

    #https://docs.teamly.one/update-a-team-12817978e0
    async def update_team(self, teamId: str,*, payload: Dict[str, Any]):
        r = Route("POST","/teams/{teamId}", teamId=teamId)
        return await self.request(r, json=payload)



    #Roles
    #https://docs.teamly.one/create-a-role-for-team-12817984e0
    async def create_role(self, teamId: str, payload: Dict[str, Any]):
        r = Route("POST","/teams/{teamId}/roles", teamId=teamId)
        return await self.request(r, json=payload)

    async def get_roles(self, teamId: str):
        r = Route("GET","/teams/{teamId}/roles", teamId=teamId)
        return await self.request(r)

    async def delete_role(self, teamId: str, roleId: str):
        r = Route("DELETE","/teams/{teamId}/roles/{roleId}", teamId=teamId, roleId=roleId)
        return await self.request(r)

    async def clone_role(self, teamId: str, roleId: str):
        r = Route("POST","/teams/{teamId}/roles/{roleId}/clone", teamId=teamId, roleId=roleId)
        return await self.request(r)

    async def update_role_priorities(self, teamId: str, payload: Dict[None,List[str]]):
        r = Route("PATCH","/teams/{teamId}/roles-priority", teamId=teamId)
        return await self.request(r, json=payload)

    async def update_role(self, teamId: str, roleId: str, payload: Dict[str, Any]):
        r = Route("POST","/teams/{teamId}/roles/{roleId}", teamId=teamId, roleId=roleId)
        return await self.request(r, json=payload)



    #User
    async def get_user(self, userId: str):
        return await self.request(Route("GET","/users/{userId}", userId=userId))

    @property
    async def get_loggedIn_user(self):
        return await self.request(Route("GET","/me"))



    #Todos
    async def get_todo_items(self, channelId: str):
        r = Route("GET","/channels/{channelId}/todo/list", channelId=channelId)
        return await self.request(r)

    async def create_todo_item(self, channelId: str, content: str):
        payload = {"content": content}
        r = Route("POST","/channels/{channelId}/todo/item", channelId=channelId)
        return await self.request(r, json=payload)

    async def delete_todo_item(self, channelId: str, todoId: str):
        r = Route("DELTE","/channels/{channelId}/todo/item/{todoId}", channelId=channelId, todoId=todoId)
        return await self.request(r)

    async def clone_todo_item(self, channelId: str, todoId: str):
        r = Route("POST","/channels/{channelId}/todo/item/{todoId}/clone", channelId=channelId, todoId=todoId)
        return await self.request(r)

    async def update_todo_item(self, channelId: str, todoId: str, content: str, completed: bool = False):
        payload = {"content": content, "changeCompleted": completed}
        r = Route("PUT","/channels/{channelId}/todo/item/{todoId}", channelId=channelId, todoId=todoId)
        return await self.request(r, json=payload)




    #Direct Message
    async def create_direct_message(self, payload: Dict[str,List[Dict[str,str]]]):
        r = Route("POST","/me/chats")
        return await self.request(r, json=payload)



    #Application
    async def get_application_submissions(self, teamId: str):
        r = Route("GET","/teams/{teamId}/applications", teamId=teamId)
        return await self.request(r)

    async def update_application_status(
        self,
        teamId: str,
        applicationId: str,
        status: Literal["accepted","rejected"]
    ):
        payload = {"status": status}
        r = Route("POST","/teams/{teamId}/applications/{applicationId}", teamId=teamId, applicationId=applicationId)
        return await self.request(r,json=payload)

    async def update_team_application_status(self, teamId: str, enable: bool):
        payload = {"enable": enable}
        r = Route("POST","/teams/{teamId}/applications/status", teamId=teamId)
        return await self.request(r, json=payload)

    async def update_team_application_questions(self, teamId: str, payload: Dict[str, Any]):
        r = Route("PATCH","/teams/{teamId}/applications", teamId=teamId)
        return await self.request(r, json=payload)

    async def get_application_by_id(self, teamId: str, applicationId: str):
        r = Route("GET","/teams/{teamId}/applications/{applicationId}", teamId=teamId, applicationId=applicationId)
        return await self.request(r)




    #Reactions
    async def get_team_custom_reactions(self, teamId: str):
        r = Route("GET","/teams/{teamId}/reactions", teamId=teamId)
        return await self.request(r)

    async def create_new_custom_reaction_for_team(self, teamId: str, name: str, payload: aiohttp.FormData):
        r = Route("POST","/teams/{teamId}/reactions", teamId=teamId)
        return await self.request(r, data=payload)

    async def update_custom_reaction(self, teamId: str, reactionId: str, name: str):
        payload = {"name": name}
        r = Route("PUT","/teams/{teamId}/reactions/{reactionId}", teamId=teamId, reactionId=reactionId)
        return await self.request(r, json=payload)

    async def delete_custom_reaction(self, teamId: str, reactionId: str):
        r = Route("DELETE","/teams/{teamId}/reactions/{reactionId}", teamId=teamId, reactionId=reactionId)
        return await self.request(r)



    #Attachments
    async def upload_attachment(self, payload: aiohttp.FormData):
        r = Route("POST","/upload")
        return await self.request(r, data=payload)




    #Voice
    async def get_credential_for_join_voice_channel(self, teamId: str, channelId, token: str = None):
        url = "/teams/{teamId}/channels/{channelId}/join"
        if token is not None:
            url += f"?token={token}"

        r = Route("GET",url, teamId=teamId, channelId=channelId)
        return await self.request(r)

    async def update_your_voice_metadata(
        self,
        teamId: str,
        channelId: str,
        isMuted: bool = True,
        isDeafened: bool = True
    ):
        payload = {"isMuted": isMuted, "isDeafened": isDeafened}
        r = Route("POST","/teams/{teamId}/channels/{channelId}/metadata", teamId=teamId, channelId=channelId)
        return await self.request(r,json=payload)

    async def leave_voice_channel(self, teamId: str, channelId: str):
        r = Route("GET","/teams/{teamId}/channels/{channelId}/leave", teamId=teamId, channelId=channelId)
        return await self.request(r)

    async def move_member(self, teamId: str, channelId: str, userId: str, fromChannelId: str):
        payload = {
            "userId": userId,
            "fromChannelId": fromChannelId
        }
        r = Route("POST","/teams/{teamId}/channels/{channelId}/move", teamId=teamId, channelId=channelId)
        return await self.request(r, json=payload)

    async def disconnect_member_from_voice(self, teamId: str, channelId: str, userId: str):
        r = Route("POST","/teams/{teamId}/channels/{channelId}/participants/{userId}/disconnect", teamId=teamId, channelId=channelId, userId=userId)
        return await self.request(r)




    #Webhooks
    async def create_webhook_message(self, webhookId: str, webhookToken: str, payload: Dict[str, Any]):
        r = Route("POST","/webhooks/{webhookId}/{webhookToken}", webhookId=webhookId, webhookToken=webhookToken)
        return await self.request(r, json=payload)

    async def webhook_for_github(self, webhookId: str, webhookToken: str):
        r = Route("POST","/webhooks/{webhookId}/{webhookToken}/github", webhookId=webhookId, webhookToken=webhookToken)
        return await self.request(r)



    #Blog
    async def get_blog_post(self, teamId: str):
        r = Route("GET","/teams/{teamId}/blogs", teamId=teamId)
        return await self.request(r)

    async def create_blog_post(self, teamId: str, payload: Dict[str, Any]):
        r = Route("POST","/teams/{teamId}/blogs", teamId=teamId)
        return await self.request(r, json=payload)

    async def delete_blog_post(self, teamId: str, blogId: str):
        r = Route("DELETE","/teams/{teamId}/blogs/{blogId}", teamId=teamId, blogId=blogId)
        return await self.request(r)



    #Category
    async def create_category(self, teamId: str, name: str):
        payload = {"name": name}
        r = Route("POST","/teams/{teamId}/categories", teamId=teamId)
        return await self.request(r,json=payload)

    async def update_category(self, teamId: str, categoryId: str, name: str):
        payload = {"name": name}
        r = Route("PUT","/teams/{teamId}/categories/{categoryId}", teamId=teamId, categoryId=categoryId)
        return await self.request(r,json=payload)

    async def update_category_role_permission(
        self,
        teamId: str,
        categoryId: str,
        roleId: str,
        allow: int = 0,
        deny: int = 0
    ):
        payload = {
            "allow": allow,
            "deny": deny
        }
        r = Route("POST","/teams/{teamId}/categories/{categoryId}/permissions/role/{roleId}", teamId=teamId, categoryId=categoryId, roleId=roleId)
        return await self.request(r,json=payload)

    async def delete_category(self, teamId: str, categoryId: str):
        r = Route("DELETE","/teams/{teamId}/categories/{categoryId}", teamId=teamId, categoryId=categoryId)
        return await self.request(r)

    async def add_channel_to_category(self, teamId: str, categoryId: str, channelId: str):
        r = Route("POST","/teams/{teamId}/categories/{categoryId}/channels/{channelId}", teamId=teamId, categoryId=categoryId, channelId=channelId)
        return await self.request(r)

    async def delete_channel_from_category(self, teamId: str, categoryId: str, channelId: str):
        r = Route("DELETE","/teams/{teamId}/categories/{categoryId}/channels/{channelId}", teamId=teamId, categoryId=categoryId, channelId=channelId)
        return await self.request(r)

    async def set_channel_priority_of_category(self, teamId: str, categoryId: str, payload: Dict[str, List[str]]):
        r = Route("POST","/teams/{teamId}/categories/{categoryId}/channels-priority", teamId=teamId, categoryId=categoryId)
        return await self.request(r, json=payload)

    async def set_team_category_priority(self, teamId: str, payload: Dict[str,List[str]]):
        r = Route("POST","/teams/{teamId}/categories-priority", teamId=teamId)
        return await self.request(r, json=payload)




    #Announcements
    async def get_announcments(self, channelId: str):
        r = Route("GET","/channels/{channelId}/announcements", channelId=channelId)
        return await self.request(r)

    async def create_announcment(self, channelId: str, payload: Dict[str, Any]):
        r = Route("POST","/channels/{channelId}/announcements", channelId=channelId)
        return await self.request(r, json=payload)

    async def delete_announcement(self, channelId: str, announcementId: str):
        r = Route("DELETE","/channels/{channelId}/announcements/{announcementId}", channelId=channelId, announcementId=announcementId)
        return await self.request(r)



'''
MIT License

Copyright (c) 2025 Fatih Kuloglu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import asyncio
import inspect
import json

from loguru import logger

from teamly.announcement import Announcement
from teamly.application import Application
from teamly.member import Member
from teamly.message import Message


from .cache import Cache
from .todo import TodoItem
from .channel import _channel_factory
from .blog import Blog
from .http import HTTPClient

from typing import Dict, Callable, Any


class ConnectionState:

    def __init__(self, dispatch: Callable[...,Any],http: HTTPClient) -> None:
        self.http: HTTPClient = http
        self.dispatch: Callable[...,Any] = dispatch

        self.parsers: Dict[str, Callable[[Any], None]]
        self.parsers = parsers = {}
        for attr, func in inspect.getmembers(self):
            if attr.startswith('parse_'):
                parsers[attr[6:].upper()] = func

        self.clear()

    def clear(self):
        self.cache: Cache = Cache(state=self)


    def parse_ready(self, data: Any):
        logger.info("Bot connected successfuly")
        asyncio.create_task(self.__setup_before_ready(data), name = "setup_ready")

    async def __setup_before_ready(self, data: Any):
        await asyncio.wait_for(self.cache.setup_cache(data=data), timeout=15)
        self.dispatch('ready')



    def parse_channel_created(self, data: Dict[str,Any]):
        factory = _channel_factory(data['channel']['type'])
        team = self.cache.get_team(teamId=data['teamId'])
        if factory:
            channel = factory(state=self,team=team, data=data['channel'])
            self.cache.add_channel(teamId=channel.team_id, channelId=channel.id, channel=channel)
            self.dispatch('channel_created', channel)

    def parse_channel_deleted(self, data: Any):
        self.cache.delete_channel(teamId=data['teamId'], channelId=data['channelId'])
        self.dispatch('channel_deleted',data)

    def parse_channel_updated(self, data: Any):
        factory = _channel_factory(data['channel']['type'])
        team = self.cache.get_team(teamId=data['teamId'])
        if factory:
            channel = factory(state=self,team=team, data=data['channel'])
            self.cache.update_channel(teamId=channel.team_id, channelId=channel.id, channel=channel)
            self.dispatch('channel_updated', channel)



    def parse_message_send(self, data: Any):
        channel = self.cache.get_channel(teamId=data['teamId'], channelId=data['channelId'])
        message = Message(state=self,channel=channel, data=data['message'])
        self.dispatch("message",message)

    def parse_message_updated(self, data: Any):
        self.dispatch("message_updated", data)

    def parse_message_deleted(self, data: Any):
        self.dispatch("message_deleted", data)

    def parse_message_reaction_added(self, data: Any):
        self.dispatch("message_reaction", data)

    def parse_message_reaction_removed(self, data: Any):
        self.dispatch("message_reaction_removed", data)



    def parse_presence_update(self, data: Any):
        self.dispatch("presence_updated",data)



    def parse_team_role_created(self, data: Any):
        self.dispatch("team_role", data)

    def parse_team_role_deleted(self, data: Any):
        self.dispatch("team_role_deleted", data)

    def parse_team_roles_updated(self, data: Any):
        self.dispatch("team_roles_updated", data)

    def parse_team_updated(self, data: Any):
        self.dispatch("team_updated", data)



    def parse_todo_item_created(self, data: Any):
        channel = self.cache.get_channel(teamId=data['teamId'], channelId=data['channelId'])
        todo_item = TodoItem(state=self,channel=channel, data=data['todo'])
        self.dispatch("todo_item", todo_item)
        print(json.dumps(data,indent=4, ensure_ascii=False))

    def parse_todo_item_deleted(self, data: Any):
        self.dispatch("todo_item_deleted", data)

    def parse_todo_item_updated(self, data: Any):
        channel = self.cache.get_channel(teamId=data['teamId'], channelId=data['channelId'])
        todo_item = TodoItem(state=self,channel=channel, data=data)
        self.dispatch("todo_item_updated", todo_item)




    def parse_user_joined_team(self, data: Any):
        member = Member._new_member(state=self, data=data['user'], teamId=data['teamId'])
        self.cache.add_member(teamId=data['teamId'], member=member)
        self.dispatch("user_joined_team")
        print('parse_user_joined_team')
        print(json.dumps(data,indent=4, ensure_ascii=False))

    def parse_user_left_team(self, data: Any):
        self.dispatch("user_left_team")
        print(json.dumps(data,indent=4, ensure_ascii=False))


    def parse_user_joined_voice_channel(self, data: Any):
        print(data)
        self.cache.voice_participants_joined(teamId=data['teamId'], channelId=data['channelId'],participantId=data['user']['id'])
        voice = self.cache.get_channel(teamId=data['teamId'], channelId=data['channelId'])
        self.dispatch("user_joinded_voice_channel",voice)

    def parse_user_left_voice_channel(self, data: Any):
        self.cache.voice_participants_leaved(teamId=data['teamId'], channelId=data['channelId'],participantId=data['user']['id'])
        voice = self.cache.get_channel(teamId=data['teamId'], channelId=data['channelId'])
        self.dispatch("user_left_voice_channel", voice)


    def parse_user_profile_updated(self, data: Any):
        self.dispatch("user_profile_updated")
        print(json.dumps(data,indent=4, ensure_ascii=False))

    def parse_user_role_added(self, data: Any):
        self.dispatch("user_role_added")
        print(json.dumps(data,indent=4, ensure_ascii=False))

    def parse_user_role_removed(self, data: Any):
        self.dispatch("user_role_removed")
        print(json.dumps(data,indent=4, ensure_ascii=False))

    def parse_user_updated_voice_metadata(self, data: Any):
        print(data)
        self.dispatch("user_updated_voice_metadata")




    def parse_blog_created(self, data: Any):
        team = self.cache.get_team(teamId=data['teamId'])
        blog = Blog(state=self,team=team, data=data)
        self.dispatch("blog", blog)

    def parse_blog_deleted(self, data: Any):
        self.dispatch("blog_deleted", data)




    def parse_categories_priority_updated(self, data: Any):
        self.dispatch("categories_priority_updated")
        print(json.dumps(data,indent=4, ensure_ascii=False))

    def parse_category_updated(self, data: Any):
        self.dispatch("category_updated")
        print(json.dumps(data,indent=4, ensure_ascii=False))

    def parse_category_deleted(self, data: Any):
        self.dispatch("category_deleted")
        print(json.dumps(data,indent=4, ensure_ascii=False))

    def parse_category_created(self, data: Any):
        self.dispatch("category")
        print(json.dumps(data,indent=4, ensure_ascii=False))

    def parse_channels_priority_updated(self, data: Any):
        self.dispatch("channels_priority_updated")
        print(json.dumps(data,indent=4, ensure_ascii=False))


    def parse_announcement_created(self, data: Any):
        channel = self.cache.get_channel(teamId=data['teamId'], channelId=data['channelId'])
        announcement = Announcement(state=self,channel=channel, data=data['announcement'])
        self.dispatch("announcement",announcement)

    def parse_announcement_deleted(self, data: Any):
        self.dispatch("announcement_deleted",data)




    def parse_application_created(self, data: Any):
        team = self.cache.get_team(teamId=data['teamId'])
        if team:
            app = Application(state=self,team=team,data=data)
            self.dispatch("application", app)

    def parse_application_updated(self, data: Any):
        team = self.cache.get_team(teamId=data['teamId'])
        if team:
            app = Application(state=self,team=team,data=data['application'])
            self.dispatch("application_updated", app)




    def parse_voice_channel_move(self, data: Any):
        self.dispatch("voice_channel_move")
        print(json.dumps(data,indent=4, ensure_ascii=False))
